
/*******************************************************************************
r.sunyear: This program is based on the module r.sun and takes the calculations of the solar geometry and solar radiation model from that module. r.sunyear calculates for a whole year the optimum inclination angle that gives the maximum annual insolation, as well as the beam, diffuse and reflected irradiation on that plane. 
(C) 2016 Copyright Thomas Huld, European Commission Joint Research Centre
e-mail: thomas.huld@jrc.ec.europa.eu
*******************************************************************************/
/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 *   Free Software Foundation, Inc.,
 *   59 Temple Place - Suite 330,
 *   Boston, MA  02111-1307, USA.
 */

/*v. 2.0 July 2002, NULL data handling, JH */
/*v. 2.1 January 2003, code optimization by JRC, JH */

#define HOURANGLE M_PI/12.
#define UNDEF	0.              /* undefined value for terrain aspect */
#define UNDEFZ   -9999.         /* undefined value for NULL */
#define SKIP	"1"
#define BIG	  1.e20
#define EPS	  1.e-4
#define LINKE	"3.0"
#define ALB	  "0.2"
#define STEP	 "0.5"
#define BSKY	  1.0
#define DSKY	  1.0
#define DIST	 "1.0"
#define NUM_PARTITIONS "1"

#define SCALING_FACTOR 150.
const double invScale = 1. / SCALING_FACTOR;

#define DISTANCE1(x1, x2, y1, y2) (sqrt((x1 - x2)*(x1 - x2) + (y1 - y2)*(y1 - y2)))

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <grass/gis.h>
#include <grass/raster.h>
/*
   #include "Vect.h"
 */
#include <grass/gprojects.h>
#include "sunradstruct.h"
#include "local_proto.h"
#include "rsunglobals.h"


const double pihalf = M_PI * 0.5;

const double pi2 = M_PI * 2.;

const double deg2rad = M_PI / 180.;

const double rad2deg = 180. / M_PI;

double fmin(double x, double y);

char *str_step;


FILE *felevin, *faspin, *flinkein, *falbedo, *flatin;

FILE *fbeam_rad, *finsol_time, *fdiff_rad, *frefl_rad;

FILE *fw;

char *elevin;

char *aspin;

char *linkein = NULL;

char *albedo = NULL;

char *latin = NULL;

char *coefbh = NULL;

char *coefdh = NULL;

char *horizon = NULL;

char *opt_angle = NULL;

char *beam_rad = NULL;

char *diff_rad = NULL;

char *refl_rad = NULL;

char *mapset = NULL;

char *per;

char *shade;

char mapname[1024];

struct Cell_head cellhd;

struct Key_Value *in_proj_info, *in_unit_info;

struct pj_info iproj;

struct pj_info oproj;


double com_declin(int);

int INPUT(double *zmax);

int INPUT_part(int offset);

int OUTGR(void);

double amax1(double, double);

double amin1(double, double);

int min(int, int);

int max(int, int);

int new_point();

void where_is_point();

void cube(int, int);

void (*func) (int, int);

void calculate(double zmax, double singleLinke, double singleAlbedo,
               struct SolarRadVar globalRadValues,
               struct GridGeometry gridGeom);
void joules2(struct SunGeometryConstDay *sunGeom,
             struct SunGeometryVarDay *sunVarGeom,
             struct SunGeometryVarSlope *sunSlopeGeom,
             struct SolarRadVar *sunRadVar,
             struct GridGeometry *gridGeom,
             unsigned char *horizonpointer,
             double latitude, double longitude);

int shd, typ, n, m, ip, jp;

float **z, **o, **s;

float **li1, **li2, **li3, **li4, **li5, **li6, **li7, **li8, **li9, **li10,
    **li11, **li12;
float **a, **la;

float **cbhr1, **cbhr2, **cbhr3, **cbhr4, **cbhr5, **cbhr6, **cbhr7, **cbhr8,
    **cbhr9, **cbhr10, **cbhr11, **cbhr12;
float **cdhr1, **cdhr2, **cdhr3, **cdhr4, **cdhr5, **cdhr6, **cdhr7, **cdhr8,
    **cdhr9, **cdhr10, **cdhr11, **cdhr12;
double stepx, stepy, stepxy, xp, yp, op, dp, xg0, xx0, yg0, yy0, deltx, delty;

double invstepx, invstepy;

int saveMemory, numPartitions = 1;

int arrayNumInt;

float **lumcl, **beam, **optang, **diff, **refl;

unsigned char *horizonarray = NULL;

int varCount_global = 0;

int bitCount_global = 0;

double xmin, xmax, ymin, ymax;

int d, day[12], tlac = 0;

double zmult = 1.0, c[12], declin[12], step, dist;

double li_max = 0., li_min = 100., al_max = 0., al_min = 1.0, la_max =
    0., la_min = 90.;
char *lt;

double horizonStep;

double ltime, declination[12];

double beam_e, diff_e, refl_e, bh, dh, rr, insol_t;

double TOLER;

inline double amax1(arg1, arg2)
     double arg1;

     double arg2;
{
    double res;

    if (arg1 >= arg2) {
        res = arg1;
    }
    else {
        res = arg2;
    }
    return res;
}


inline double amin1(arg1, arg2)
     double arg1;

     double arg2;
{
    double res;

    if (arg1 <= arg2) {
        res = arg1;
    }
    else {
        res = arg2;
    }
    return res;
}

void cube(int jmin, int imin)
{
}


int main(int argc, char *argv[])
{
    int i;

    FILE *fp;

    double zmax;

    double singleLinke;

    double singleAlbedo;

    struct GridGeometry gridGeom;

    struct SolarRadVar globalRadValues;


    struct GModule *module;

    struct
    {
        struct Option *elevin, *aspin, *linkein, *lin, *albedo, *alb, *latin,
            *coefbh, *coefdh, *opt_angle, *beam_rad, *diff_rad, *refl_rad,
            *day, *step, *declin, *ltime, *dist, *horizon, *horizonstep,
            *numPartitions;
    } parm;

    struct
    {
        struct Flag *shade;
    } flag;



    parm.lin = NULL;
    parm.alb = NULL;


    G_gisinit(argv[0]);
    module = G_define_module();

    module->description =
        "Computes direct (beam), diffuse and reflected solar irradiation raster "
        "maps for given day, latitude, surface and atmospheric conditions. Solar "
        "parameters (e.g. sunrise, sunset times, declination, extraterrestrial "
        "irradiance, daylight length) are saved in a local text file. "
        "Alternatively, a local time can be specified to compute solar "
        "incidence angle and/or irradiance raster maps. The shadowing effect of "
        "the topography is optionally incorporated. ";

    G_get_set_window(&cellhd);
    gridGeom.stepx = cellhd.ew_res;
    gridGeom.stepy = cellhd.ns_res;
    invstepx = 1. / stepx;
    invstepy = 1. / stepy;
    n /*n_cols */  = cellhd.cols;
    m /*n_rows */  = cellhd.rows;
    xmin = cellhd.west;
    ymin = cellhd.south;
    xmax = cellhd.east;
    ymax = cellhd.north;
    gridGeom.deltx = fabs(cellhd.east - cellhd.west);
    gridGeom.delty = fabs(cellhd.north - cellhd.south);

    parm.elevin = G_define_option();
    parm.elevin->key = "elevation";
    parm.elevin->type = TYPE_STRING;
    parm.elevin->required = YES;
    parm.elevin->gisprompt = "old,cell,raster";
    parm.elevin->description = "Name of the elevation raster file";

    parm.aspin = G_define_option();
    parm.aspin->key = "aspect";
    parm.aspin->type = TYPE_STRING;
    parm.aspin->required = YES;
    parm.aspin->gisprompt = "old,cell,raster";
    parm.aspin->description = "Name of the aspect raster file";

    parm.linkein = G_define_option();
    parm.linkein->key = "linke";
    parm.linkein->type = TYPE_STRING;
    parm.linkein->required = NO;
    parm.linkein->gisprompt = "old,cell,raster";
    parm.linkein->description =
        "Name of the Linke's turbidity coefficient raster file";

    parm.lin = G_define_option();
    parm.lin->key = "lin";
    parm.lin->type = TYPE_DOUBLE;
    parm.lin->answer = LINKE;
    parm.lin->required = NO;
    parm.lin->description =
        "A single value of the Linke's turbidity coefficient";

    parm.albedo = G_define_option();
    parm.albedo->key = "albedo";
    parm.albedo->type = TYPE_STRING;
    parm.albedo->required = NO;
    parm.albedo->gisprompt = "old,cell,raster";
    parm.albedo->description = "Name of the albedo coefficient raster file";

    parm.alb = G_define_option();
    parm.alb->key = "albedo_value";
    parm.alb->type = TYPE_DOUBLE;
    parm.alb->answer = ALB;
    parm.alb->required = NO;
    parm.alb->description = "A single value of the albedo coefficient";

    parm.latin = G_define_option();
    parm.latin->key = "latitude";
    parm.latin->type = TYPE_STRING;
    parm.latin->required = NO;
    parm.latin->gisprompt = "old,cell,raster";
    parm.latin->description = "Name of the latitude raster file";


    parm.coefbh = G_define_option();
    parm.coefbh->key = "coefbh";
    parm.coefbh->type = TYPE_STRING;
    parm.coefbh->required = NO;
    parm.coefbh->gisprompt = "old,cell,raster";
    parm.coefbh->description =
        " The real-sky beam radiation coefficient file";

    parm.coefdh = G_define_option();
    parm.coefdh->key = "coefdh";
    parm.coefdh->type = TYPE_STRING;
    parm.coefdh->required = NO;
    parm.coefdh->gisprompt = "old,cell,raster";
    parm.coefdh->description =
        "The real-sky diffuse radiation coefficient file";


    parm.horizon = G_define_option();
    parm.horizon->key = "horizon_basename";
    parm.horizon->type = TYPE_STRING;
    parm.horizon->required = NO;
    parm.horizon->gisprompt = "old,cell,raster";
    parm.horizon->description = "The horizon information file prefix";

    parm.horizonstep = G_define_option();
    parm.horizonstep->key = "horizon_step";
    parm.horizonstep->type = TYPE_DOUBLE;
    parm.horizonstep->required = NO;
    parm.horizonstep->description =
        "Angle step size for the horizon information (degrees)";


    parm.opt_angle = G_define_option();
    parm.opt_angle->key = "opt_angle";
    parm.opt_angle->type = TYPE_STRING;
    parm.opt_angle->required = NO;
    parm.opt_angle->gisprompt = "old,cell,raster";
    parm.opt_angle->description =
        "Output optimal angle for solar irradiance, file (raster)";

    parm.beam_rad = G_define_option();
    parm.beam_rad->key = "beam_rad";
    parm.beam_rad->type = TYPE_STRING;
    parm.beam_rad->required = NO;
    parm.beam_rad->gisprompt = "old,cell,raster";
    parm.beam_rad->description =
        "Output direct (beam) irradiance/irradiation file (raster)";

    parm.diff_rad = G_define_option();
    parm.diff_rad->key = "diff_rad";
    parm.diff_rad->type = TYPE_STRING;
    parm.diff_rad->required = NO;
    parm.diff_rad->gisprompt = "old,cell,raster";
    parm.diff_rad->description =
        "Output diffuse irradiance/irradiation file (raster)";

    parm.refl_rad = G_define_option();
    parm.refl_rad->key = "refl_rad";
    parm.refl_rad->type = TYPE_STRING;
    parm.refl_rad->required = NO;
    parm.refl_rad->gisprompt = "old,cell,raster";
    parm.refl_rad->description =
        "Output reflected irradiance/irradiation file (raster)";

    parm.day = G_define_option();
    parm.day->key = "day";
    parm.day->type = TYPE_STRING;
    parm.day->required = YES;
    parm.day->description = "File containing day numbers (1-365)";

    parm.step = G_define_option();
    parm.step->key = "step";
    parm.step->type = TYPE_DOUBLE;
    parm.step->answer = STEP;
    parm.step->required = NO;
    parm.step->description = "Time step computing all-day radiation";

    parm.declin = G_define_option();
    parm.declin->key = "declin";
    parm.declin->type = TYPE_STRING;
    parm.declin->required = NO;
    parm.declin->description =
        "Required declination values (overriding the internal value)";

    parm.ltime = G_define_option();
    parm.ltime->key = "time";
    parm.ltime->type = TYPE_DOUBLE;
    /*                parm.ltime->answer = TIME; */
    parm.ltime->required = NO;
    parm.ltime->description = "Local time";

    parm.dist = G_define_option();
    parm.dist->key = "dist";
    parm.dist->type = TYPE_DOUBLE;
    parm.dist->answer = DIST;
    parm.dist->required = NO;
    parm.dist->description = "Sampling distance step coefficient (0.5-1.5)";

    flag.shade = G_define_flag();
    flag.shade->key = 's';
    flag.shade->description =
        "Do you want to incorporate the shadowing effect of terrain (y/n)";


    parm.numPartitions = G_define_option();
    parm.numPartitions->key = "numpartitions";
    parm.numPartitions->type = TYPE_INTEGER;
    parm.numPartitions->answer = NUM_PARTITIONS;
    parm.numPartitions->required = NO;
    parm.numPartitions->description =
        "Read the input files in this number of chunks";

    if (G_parser(argc, argv))
        exit(1);


    elevin = parm.elevin->answer;
    aspin = parm.aspin->answer;
    linkein = parm.linkein->answer;
    albedo = parm.albedo->answer;
    latin = parm.latin->answer;
    coefbh = parm.coefbh->answer;
    coefdh = parm.coefdh->answer;
    horizon = parm.horizon->answer;
    setUseHorizonData(horizon != NULL);
    opt_angle = parm.opt_angle->answer;
    beam_rad = parm.beam_rad->answer;
    diff_rad = parm.diff_rad->answer;
    refl_rad = parm.refl_rad->answer;

    fp = fopen(parm.day->answer, "r");
    for (i = 0; i < 12; i++) {
        fscanf(fp, "%d", day + i);
    }
    fclose(fp);
    sscanf(parm.step->answer, "%lf", &step);
    if (parm.linkein->answer == NULL) {
        sscanf(parm.lin->answer, "%lf", &singleLinke);
    }
    if (parm.albedo->answer == NULL)
        sscanf(parm.alb->answer, "%lf", &singleAlbedo);

    if (parm.horizonstep->answer != NULL) {
        str_step = parm.horizonstep->answer;
        setUseShadow(1);
        if (sscanf(parm.horizonstep->answer, "%lf", &horizonStep) != 1) {
            G_fatal_error("Error reading horizon step size");

        }
        setHorizonInterval(deg2rad * horizonStep);
    }



    sscanf(parm.dist->answer, "%lf", &dist);


    if (parm.numPartitions->answer != NULL) {
        sscanf(parm.numPartitions->answer, "%d", &numPartitions);
        if (useShadow() && (!useHorizonData()) && (numPartitions != 1)) {
            /* If you calculate shadows on the fly, the number of partitions
               must be one.
             */
            G_fatal_error
                ("If you use -s and no horizon rasters, numpartitions must be =1");

        }

    }

    gridGeom.stepxy = dist * 0.5 * (gridGeom.stepx + gridGeom.stepy);
    TOLER = gridGeom.stepxy * EPS;

    if (parm.declin->answer == NULL) {
        for (i = 0; i < 12; i++) {
            declination[i] = com_declin(day[i]);
        }
    }
    else {
        fp = fopen(parm.declin->answer, "r");
        for (i = 0; i < 12; i++) {
            if (fscanf(fp, "%lf", declin + i) != 1)
                G_fatal_error("Error reading declination values");
            declination[i] = -declin[i];
        }
        fclose(fp);
    }



    if (saveMemory && useShadow() && (!useHorizonData())) {
        printf
            ("If you want to save memory and to use shadows, you must calculate them beforehand.\n");
        exit(1);

    }



    if (useShadow()) {
        /*        Number of bytes holding the horizon information */
        arrayNumInt = (int)(360. / horizonStep);

    }
    else {
        arrayNumInt = 1;
    }


    /*
       tim = (timo - 12) * 15;
     */
    /* converting to degrees */
    /* Jenco (12-lum_time) * 15 */
    /*
       if (tim < 0)
       tim += 360;
       tim = deg2rad* tim;
     */
    /* conv. to radians */

    /* Set up parameters for projection to lat/long  */



    if ((in_proj_info = G_get_projinfo()) == NULL)
        G_fatal_error
            ("Can't get projection info of current location: please set latitude via 'lat' or 'latin' option!");

    if ((in_unit_info = G_get_projunits()) == NULL)
        G_fatal_error("Can't get projection units of current location");

    if (pj_get_kv(&iproj, in_proj_info, in_unit_info) < 0)
        G_fatal_error("Can't get projection key values of current location");

    /* Set output projection to latlong w/ same ellipsoid */
    oproj.zone = 0;
    oproj.meters = 1.;
    sprintf(oproj.proj, "ll");
    if ((oproj.pj = pj_latlong_from_proj(iproj.pj)) == NULL)
        G_fatal_error("Unable to set up lat/long projection parameters");


    G_free_key_value(in_proj_info);
    G_free_key_value(in_unit_info);


/**********end of parser - ******************************/

    INPUT(&zmax);
    calculate(zmax, singleLinke, singleAlbedo, globalRadValues, gridGeom);
    OUTGR();

    return 1;
}


int INPUT(double *zmax)
{
    FCELL *cell1, *cell2;

    FCELL *cell3, *cell5 = NULL, *cell6 = NULL;

    int fd1, fd2, fd5 = -1, fd6 = -1, row, row_rev;

    int l, i, j;

    cell1 = Rast_allocate_f_buf();
    cell2 = Rast_allocate_f_buf();
    cell3 = Rast_allocate_f_buf();

    z = (float **)malloc(sizeof(float *) * (m));
    o = (float **)malloc(sizeof(float *) * (m));
    s = (float **)malloc(sizeof(float *) * (m));

    for (l = 0; l < m; l++) {
        z[l] = (float *)malloc(sizeof(float) * (n));
        o[l] = (float *)malloc(sizeof(float) * (n));
        s[l] = (float *)malloc(sizeof(float) * (n));

    }

    fd1 = Rast_open_old(elevin, mapset);


    fd2 = Rast_open_old(aspin, mapset);


    if (albedo != NULL) {
        cell5 = Rast_allocate_f_buf();
        a = (float **)malloc(sizeof(float *) * (m));
        for (l = 0; l < m; l++)
            a[l] = (float *)malloc(sizeof(float) * (n));


        fd5 = Rast_open_old(albedo, mapset);
    }

    if (latin != NULL) {
        cell6 = Rast_allocate_f_buf();
        la = (float **)malloc(sizeof(float *) * (m));
        for (l = 0; l < m; l++)
            la[l] = (float *)malloc(sizeof(float) * (n));


        fd6 = Rast_open_old(latin, mapset);
    }




    for (row = 0; row < m; row++) {
        Rast_get_f_row(fd1, cell1, row);
        Rast_get_f_row(fd2, cell2, row);
        if (albedo != NULL)
            Rast_get_f_row(fd5, cell5, row);
        if (latin != NULL)
            Rast_get_f_row(fd6, cell6, row);



        row_rev = m - row - 1;

        for (j = 0; j < n; j++) {
            if (!Rast_is_f_null_value(cell1 + j))
                z[row_rev][j] = (float)cell1[j];
            else
                z[row_rev][j] = UNDEFZ;

            if (!Rast_is_f_null_value(cell2 + j))
                o[row_rev][j] = (float)cell2[j];
            else
                o[row_rev][j] = UNDEFZ;

            if (!Rast_is_f_null_value(cell3 + j))
                s[row_rev][j] = (float)cell3[j];
            else
                s[row_rev][j] = UNDEFZ;



            if (albedo != NULL) {
                if (!Rast_is_f_null_value(cell5 + j))
                    a[row_rev][j] = (float)cell5[j];
                else
                    a[row_rev][j] = UNDEFZ;
            }

            if (latin != NULL) {
                if (!Rast_is_f_null_value(cell6 + j))
                    la[row_rev][j] = (float)cell6[j];
                else
                    la[row_rev][j] = UNDEFZ;
            }


        }
    }
    Rast_close(fd1);
    Rast_close(fd2);
    if (albedo != NULL)
        Rast_close(fd5);
    if (latin != NULL)
        Rast_close(fd6);

/*******transformation of angles from 0 to east counterclock
		to 0 to north clocwise, for ori=0 upslope flowlines
		turn the orientation 2*M_PI ************/

    /* needs to be eliminated */

    /*for (i = 0; i < m; ++i) */
    for (i = 0; i < m; i++) {
        for (j = 0; j < n; j++) {
            *zmax = amax1(*zmax, z[i][j]);
            if (o[i][j] != 0.) {
                if (o[i][j] < 90.)
                    o[i][j] = 90. - o[i][j];
                else
                    o[i][j] = 450. - o[i][j];
                /*   printf("o,z = %d  %d i,j, %d %d \n", o[i][j],z[i][j],i,j); */

                if (z[i][j] == UNDEFZ || o[i][j] == UNDEFZ ||
                    s[i][j] == UNDEFZ)
                    z[i][j] = UNDEFZ;
                if (albedo != NULL && a[i][j] == UNDEFZ)
                    z[i][j] = UNDEFZ;
                if (latin != NULL && la[i][j] == UNDEFZ)
                    z[i][j] = UNDEFZ;
            }

        }
    }

    return 1;
}


int INPUT_part(int offset)
{
    int mon;

    int numRows;

    int finalRow, rowrevoffset;

    int decimals;

    char fileName[80];

    static FCELL *licell[12] = { NULL, NULL, NULL, NULL, NULL, NULL,
        NULL, NULL, NULL, NULL, NULL, NULL
    };
    static FCELL *rast1[12] = { NULL, NULL, NULL, NULL, NULL, NULL,
        NULL, NULL, NULL, NULL, NULL, NULL
    };
    static FCELL *rast2[12] = { NULL, NULL, NULL, NULL, NULL, NULL,
        NULL, NULL, NULL, NULL, NULL, NULL
    };
    static FCELL **horizonbuf;

    char *horizonpointer;

    int row, row_rev;

    static int *fd_shad = NULL;

    int lif[12], fr1[12], fr2[12];

    int l, i, j;

    char *shad_filename;

    double angle_deg;

    finalRow = m - offset - m / numPartitions;
    if (finalRow < 0) {
        finalRow = 0;
    }
    numRows = m / numPartitions;


    if (linkein != NULL) {
        if (licell[0] == NULL) {
            /* First call, allocate all necessary memory */
            for (i = 0; i < 12; i++) {
                licell[i] = Rast_allocate_f_buf();
            }
            li1 = (float **)malloc(sizeof(float *) * numRows);
            li2 = (float **)malloc(sizeof(float *) * numRows);
            li3 = (float **)malloc(sizeof(float *) * numRows);
            li4 = (float **)malloc(sizeof(float *) * numRows);
            li5 = (float **)malloc(sizeof(float *) * numRows);
            li6 = (float **)malloc(sizeof(float *) * numRows);
            li7 = (float **)malloc(sizeof(float *) * numRows);
            li8 = (float **)malloc(sizeof(float *) * numRows);
            li9 = (float **)malloc(sizeof(float *) * numRows);
            li10 = (float **)malloc(sizeof(float *) * numRows);
            li11 = (float **)malloc(sizeof(float *) * numRows);
            li12 = (float **)malloc(sizeof(float *) * numRows);
            for (l = 0; l < numRows; l++) {
                li1[l] = (float *)malloc(sizeof(float) * (n));
                li2[l] = (float *)malloc(sizeof(float) * (n));
                li3[l] = (float *)malloc(sizeof(float) * (n));
                li4[l] = (float *)malloc(sizeof(float) * (n));
                li5[l] = (float *)malloc(sizeof(float) * (n));
                li6[l] = (float *)malloc(sizeof(float) * (n));
                li7[l] = (float *)malloc(sizeof(float) * (n));
                li8[l] = (float *)malloc(sizeof(float) * (n));
                li9[l] = (float *)malloc(sizeof(float) * (n));
                li10[l] = (float *)malloc(sizeof(float) * (n));
                li11[l] = (float *)malloc(sizeof(float) * (n));
                li12[l] = (float *)malloc(sizeof(float) * (n));
            }
        }
        for (i = 0; i < 12; i++) {
            sprintf(fileName, "%s%02d", linkein, i + 1);
            lif[i] = Rast_open_old(fileName, mapset);

        }
    }



    if (coefbh != NULL) {
        if (rast1[0] == NULL) {
            /* First call, allocate all necessary memory */
            for (i = 0; i < 12; i++) {
                rast1[i] = Rast_allocate_f_buf();
            }
            cbhr1 = (float **)malloc(sizeof(float *) * numRows);
            cbhr2 = (float **)malloc(sizeof(float *) * numRows);
            cbhr3 = (float **)malloc(sizeof(float *) * numRows);
            cbhr4 = (float **)malloc(sizeof(float *) * numRows);
            cbhr5 = (float **)malloc(sizeof(float *) * numRows);
            cbhr6 = (float **)malloc(sizeof(float *) * numRows);
            cbhr7 = (float **)malloc(sizeof(float *) * numRows);
            cbhr8 = (float **)malloc(sizeof(float *) * numRows);
            cbhr9 = (float **)malloc(sizeof(float *) * numRows);
            cbhr10 = (float **)malloc(sizeof(float *) * numRows);
            cbhr11 = (float **)malloc(sizeof(float *) * numRows);
            cbhr12 = (float **)malloc(sizeof(float *) * numRows);

            for (l = 0; l < numRows; l++) {
                cbhr1[l] = (float *)malloc(sizeof(float) * (n));
                cbhr2[l] = (float *)malloc(sizeof(float) * (n));
                cbhr3[l] = (float *)malloc(sizeof(float) * (n));
                cbhr4[l] = (float *)malloc(sizeof(float) * (n));
                cbhr5[l] = (float *)malloc(sizeof(float) * (n));
                cbhr6[l] = (float *)malloc(sizeof(float) * (n));
                cbhr7[l] = (float *)malloc(sizeof(float) * (n));
                cbhr8[l] = (float *)malloc(sizeof(float) * (n));
                cbhr9[l] = (float *)malloc(sizeof(float) * (n));
                cbhr10[l] = (float *)malloc(sizeof(float) * (n));
                cbhr11[l] = (float *)malloc(sizeof(float) * (n));
                cbhr12[l] = (float *)malloc(sizeof(float) * (n));
            }
        }

        for (i = 0; i < 12; i++) {
            sprintf(fileName, "%s%02d", coefbh, i + 1);
            fr1[i] = Rast_open_old(fileName, mapset);

        }

    }

    if (coefdh != NULL) {
        if (rast2[0] == NULL) {
            /* First call, allocate all necessary memory */
            for (i = 0; i < 12; i++) {
                rast2[i] = Rast_allocate_f_buf();
            }
            cdhr1 = (float **)malloc(sizeof(float *) * numRows);
            cdhr2 = (float **)malloc(sizeof(float *) * numRows);
            cdhr3 = (float **)malloc(sizeof(float *) * numRows);
            cdhr4 = (float **)malloc(sizeof(float *) * numRows);
            cdhr5 = (float **)malloc(sizeof(float *) * numRows);
            cdhr6 = (float **)malloc(sizeof(float *) * numRows);
            cdhr7 = (float **)malloc(sizeof(float *) * numRows);
            cdhr8 = (float **)malloc(sizeof(float *) * numRows);
            cdhr9 = (float **)malloc(sizeof(float *) * numRows);
            cdhr10 = (float **)malloc(sizeof(float *) * numRows);
            cdhr11 = (float **)malloc(sizeof(float *) * numRows);
            cdhr12 = (float **)malloc(sizeof(float *) * numRows);

            for (l = 0; l < numRows; l++) {
                cdhr1[l] = (float *)malloc(sizeof(float) * (n));
                cdhr2[l] = (float *)malloc(sizeof(float) * (n));
                cdhr3[l] = (float *)malloc(sizeof(float) * (n));
                cdhr4[l] = (float *)malloc(sizeof(float) * (n));
                cdhr5[l] = (float *)malloc(sizeof(float) * (n));
                cdhr6[l] = (float *)malloc(sizeof(float) * (n));
                cdhr7[l] = (float *)malloc(sizeof(float) * (n));
                cdhr8[l] = (float *)malloc(sizeof(float) * (n));
                cdhr9[l] = (float *)malloc(sizeof(float) * (n));
                cdhr10[l] = (float *)malloc(sizeof(float) * (n));
                cdhr11[l] = (float *)malloc(sizeof(float) * (n));
                cdhr12[l] = (float *)malloc(sizeof(float) * (n));
            }
        }
        for (i = 0; i < 12; i++) {
            sprintf(fileName, "%s%02d", coefdh, i + 1);
            fr2[i] = Rast_open_old(fileName, mapset);

        }
    }



    if (useHorizonData()) {
        if (horizonarray == NULL) {
            horizonarray =
                (unsigned char *)malloc(sizeof(char) * arrayNumInt * numRows *
                                        n);

            horizonbuf = (FCELL **) malloc(sizeof(FCELL *) * arrayNumInt);
            fd_shad = (int *)malloc(sizeof(int) * arrayNumInt);
        }
        decimals = G_get_num_decimals(str_step);
        angle_deg = 0;
        for (i = 0; i < arrayNumInt; i++) {
            horizonbuf[i] = Rast_allocate_f_buf();
            shad_filename = G_generate_basename(horizon, angle_deg,
                                                3, decimals);
            fd_shad[i] = Rast_open_old(shad_filename, "");
            angle_deg += horizonStep;
            G_free(shad_filename);
        }


        if (horizon != NULL) {


            for (i = 0; i < arrayNumInt; i++) {
                for (row = m - offset - 1; row >= finalRow; row--) {

                    row_rev = m - row - 1;
                    rowrevoffset = row_rev - offset;
                    Rast_get_f_row(fd_shad[i], horizonbuf[i], row);
                    horizonpointer =
                        horizonarray + arrayNumInt * n * rowrevoffset;
                    for (j = 0; j < n; j++) {

                        horizonpointer[i] = (char)(rint(SCALING_FACTOR
                                                        *
                                                        fmin(horizonbuf[i][j],
                                                             256 *
                                                             invScale)));
                        horizonpointer += arrayNumInt;

                    }
                }
            }


        }

    }

    for (row = m - offset - 1; row >= finalRow; row--) {
        if (linkein != NULL) {
            for (i = 0; i < 12; i++) {
                Rast_get_f_row(lif[i], licell[i], row);
            }
        }
        if (coefbh != NULL) {
            for (i = 0; i < 12; i++) {
                Rast_get_f_row(fr1[i], rast1[i], row);
            }
        }
        if (coefdh != NULL) {
            for (i = 0; i < 12; i++) {
                Rast_get_f_row(fr2[i], rast2[i], row);
            }
        }




        row_rev = m - row - 1;
        rowrevoffset = row_rev - offset;
        for (j = 0; j < n; j++) {


            if (linkein != NULL) {
                if (!Rast_is_f_null_value(licell[0] + j))
                    li1[rowrevoffset][j] = (float)licell[0][j];
                else
                    li1[rowrevoffset][j] = UNDEFZ;
                if (!Rast_is_f_null_value(licell[1] + j))
                    li2[rowrevoffset][j] = (float)licell[1][j];
                else
                    li2[rowrevoffset][j] = UNDEFZ;
                if (!Rast_is_f_null_value(licell[2] + j))
                    li3[rowrevoffset][j] = (float)licell[2][j];
                else
                    li3[rowrevoffset][j] = UNDEFZ;
                if (!Rast_is_f_null_value(licell[3] + j))
                    li4[rowrevoffset][j] = (float)licell[3][j];
                else
                    li4[rowrevoffset][j] = UNDEFZ;
                if (!Rast_is_f_null_value(licell[4] + j))
                    li5[rowrevoffset][j] = (float)licell[4][j];
                else
                    li5[rowrevoffset][j] = UNDEFZ;
                if (!Rast_is_f_null_value(licell[5] + j))
                    li6[rowrevoffset][j] = (float)licell[5][j];
                else
                    li6[rowrevoffset][j] = UNDEFZ;
                if (!Rast_is_f_null_value(licell[6] + j))
                    li7[rowrevoffset][j] = (float)licell[6][j];
                else
                    li7[rowrevoffset][j] = UNDEFZ;
                if (!Rast_is_f_null_value(licell[7] + j))
                    li8[rowrevoffset][j] = (float)licell[7][j];
                else
                    li8[rowrevoffset][j] = UNDEFZ;
                if (!Rast_is_f_null_value(licell[8] + j))
                    li9[rowrevoffset][j] = (float)licell[8][j];
                else
                    li9[rowrevoffset][j] = UNDEFZ;
                if (!Rast_is_f_null_value(licell[9] + j))
                    li10[rowrevoffset][j] = (float)licell[9][j];
                else
                    li10[rowrevoffset][j] = UNDEFZ;
                if (!Rast_is_f_null_value(licell[10] + j))
                    li11[rowrevoffset][j] = (float)licell[10][j];
                else
                    li11[rowrevoffset][j] = UNDEFZ;
                if (!Rast_is_f_null_value(licell[11] + j))
                    li12[rowrevoffset][j] = (float)licell[11][j];
                else
                    li12[rowrevoffset][j] = UNDEFZ;


            }

            if (coefbh != NULL) {
                if (!Rast_is_f_null_value(rast1[0] + j))
                    cbhr1[rowrevoffset][j] = (float)rast1[0][j];
                else
                    cbhr1[rowrevoffset][j] = UNDEFZ;
                if (!Rast_is_f_null_value(rast1[1] + j))
                    cbhr2[rowrevoffset][j] = (float)rast1[1][j];
                else
                    cbhr2[rowrevoffset][j] = UNDEFZ;
                if (!Rast_is_f_null_value(rast1[2] + j))
                    cbhr3[rowrevoffset][j] = (float)rast1[2][j];
                else
                    cbhr3[rowrevoffset][j] = UNDEFZ;
                if (!Rast_is_f_null_value(rast1[3] + j))
                    cbhr4[rowrevoffset][j] = (float)rast1[3][j];
                else
                    cbhr4[rowrevoffset][j] = UNDEFZ;
                if (!Rast_is_f_null_value(rast1[4] + j))
                    cbhr5[rowrevoffset][j] = (float)rast1[4][j];
                else
                    cbhr5[rowrevoffset][j] = UNDEFZ;
                if (!Rast_is_f_null_value(rast1[5] + j))
                    cbhr6[rowrevoffset][j] = (float)rast1[5][j];
                else
                    cbhr6[rowrevoffset][j] = UNDEFZ;
                if (!Rast_is_f_null_value(rast1[6] + j))
                    cbhr7[rowrevoffset][j] = (float)rast1[6][j];
                else
                    cbhr7[rowrevoffset][j] = UNDEFZ;
                if (!Rast_is_f_null_value(rast1[7] + j))
                    cbhr8[rowrevoffset][j] = (float)rast1[7][j];
                else
                    cbhr8[rowrevoffset][j] = UNDEFZ;
                if (!Rast_is_f_null_value(rast1[8] + j))
                    cbhr9[rowrevoffset][j] = (float)rast1[8][j];
                else
                    cbhr9[rowrevoffset][j] = UNDEFZ;
                if (!Rast_is_f_null_value(rast1[9] + j))
                    cbhr10[rowrevoffset][j] = (float)rast1[9][j];
                else
                    cbhr10[rowrevoffset][j] = UNDEFZ;
                if (!Rast_is_f_null_value(rast1[10] + j))
                    cbhr11[rowrevoffset][j] = (float)rast1[10][j];
                else
                    cbhr11[rowrevoffset][j] = UNDEFZ;
                if (!Rast_is_f_null_value(rast1[11] + j))
                    cbhr12[rowrevoffset][j] = (float)rast1[11][j];
                else
                    cbhr12[rowrevoffset][j] = UNDEFZ;


            }


            if (coefdh != NULL) {
                if (!Rast_is_f_null_value(rast2[0] + j))
                    cdhr1[rowrevoffset][j] = (float)rast2[0][j];
                else
                    cdhr1[rowrevoffset][j] = UNDEFZ;
                if (!Rast_is_f_null_value(rast2[1] + j))
                    cdhr2[rowrevoffset][j] = (float)rast2[1][j];
                else
                    cdhr2[rowrevoffset][j] = UNDEFZ;
                if (!Rast_is_f_null_value(rast2[2] + j))
                    cdhr3[rowrevoffset][j] = (float)rast2[2][j];
                else
                    cdhr3[rowrevoffset][j] = UNDEFZ;
                if (!Rast_is_f_null_value(rast2[3] + j))
                    cdhr4[rowrevoffset][j] = (float)rast2[3][j];
                else
                    cdhr4[rowrevoffset][j] = UNDEFZ;
                if (!Rast_is_f_null_value(rast2[4] + j))
                    cdhr5[rowrevoffset][j] = (float)rast2[4][j];
                else
                    cdhr5[rowrevoffset][j] = UNDEFZ;
                if (!Rast_is_f_null_value(rast2[5] + j))
                    cdhr6[rowrevoffset][j] = (float)rast2[5][j];
                else
                    cdhr6[rowrevoffset][j] = UNDEFZ;
                if (!Rast_is_f_null_value(rast2[6] + j))
                    cdhr7[rowrevoffset][j] = (float)rast2[6][j];
                else
                    cdhr7[rowrevoffset][j] = UNDEFZ;
                if (!Rast_is_f_null_value(rast2[7] + j))
                    cdhr8[rowrevoffset][j] = (float)rast2[7][j];
                else
                    cdhr8[rowrevoffset][j] = UNDEFZ;
                if (!Rast_is_f_null_value(rast2[8] + j))
                    cdhr9[rowrevoffset][j] = (float)rast2[8][j];
                else
                    cdhr9[rowrevoffset][j] = UNDEFZ;
                if (!Rast_is_f_null_value(rast2[9] + j))
                    cdhr10[rowrevoffset][j] = (float)rast2[9][j];
                else
                    cdhr10[rowrevoffset][j] = UNDEFZ;
                if (!Rast_is_f_null_value(rast2[10] + j))
                    cdhr11[rowrevoffset][j] = (float)rast2[10][j];
                else
                    cdhr11[rowrevoffset][j] = UNDEFZ;
                if (!Rast_is_f_null_value(rast2[11] + j))
                    cdhr12[rowrevoffset][j] = (float)rast2[11][j];
                else
                    cdhr12[rowrevoffset][j] = UNDEFZ;


            }




        }                       /* End for loop over columns */
    }                           /* End for loop over rows */


    if (linkein != NULL) {
        for (i = 0; i < 12; i++) {
            Rast_close(lif[i]);
        }
    }
    if (coefbh != NULL) {
        for (i = 0; i < 12; i++) {
            Rast_close(fr1[i]);
        }
    }


    if (coefdh != NULL) {
        for (i = 0; i < 12; i++) {
            Rast_close(fr2[i]);
        }
    }



    if (useHorizonData()) {
        for (i = 0; i < arrayNumInt; i++) {
            Rast_close(fd_shad[i]);
            G_free(horizonbuf[i]);
        }
    }

    return 1;
}



int OUTGR(void)
{
    FCELL *cell8 = NULL, *cell9 = NULL, *cell10 = NULL, *cell11 =
        NULL, *optcell = NULL;
    int fd8 = -1, fd9 = -1, fd10 = -1, fd11 = -1, fd12 = -1;

    int i, iarc, j;

    char msg[100];

    if (opt_angle != NULL) {
        optcell = Rast_allocate_f_buf();
        fd12 = Rast_open_fp_new(opt_angle);
        if (fd12 < 0) {
            G_fatal_error("unable to create raster map %s", opt_angle);
            exit(1);
        }
    }

    if (beam_rad != NULL) {
        cell8 = Rast_allocate_f_buf();
        fd8 = Rast_open_fp_new(beam_rad);
        if (fd8 < 0) {
            G_fatal_error("unable to create raster map %s", beam_rad);
            exit(1);
        }
    }



    if (diff_rad != NULL) {
        cell9 = Rast_allocate_f_buf();
        fd9 = Rast_open_fp_new(diff_rad);
        if (fd9 < 0) {
            G_fatal_error("unable to create raster map %s", diff_rad);
        }
    }

    if (refl_rad != NULL) {
        cell10 = Rast_allocate_f_buf();
        fd10 = Rast_open_fp_new(refl_rad);
        if (fd10 < 0) {
            G_fatal_error("unable to create raster map %s", refl_rad);
        }
    }


    G_set_window(&cellhd);

    if (m != Rast_window_rows()) {
        fprintf(stderr, "OOPS: rows changed from %d to %d\n", m,
                Rast_window_rows());
        exit(1);
    }
    if (n != Rast_window_cols()) {
        fprintf(stderr, "OOPS: cols changed from %d to %d\n", n,
                Rast_window_cols());
        exit(1);
    }

    for (iarc = 0; iarc < m; iarc++) {
        i = m - iarc - 1;
        if (opt_angle != NULL) {
            for (j = 0; j < n; j++) {
                if (optang[i][j] == UNDEFZ)
                    Rast_set_f_null_value(optcell + j, 1);
                else
                    optcell[j] = (FCELL) optang[i][j];

            }
            Rast_put_f_row(fd12, optcell);
        }

        if (beam_rad != NULL) {
            for (j = 0; j < n; j++) {
                if (beam[i][j] == UNDEFZ)
                    Rast_set_f_null_value(cell8 + j, 1);
                else
                    cell8[j] = (FCELL) beam[i][j];

            }
            Rast_put_f_row(fd8, cell8);
        }


        if (diff_rad != NULL) {
            for (j = 0; j < n; j++) {
                if (diff[i][j] == UNDEFZ)
                    Rast_set_f_null_value(cell9 + j, 1);
                else
                    cell9[j] = (FCELL) diff[i][j];
            }
            Rast_put_f_row(fd9, cell9);
        }

        if (refl_rad != NULL) {
            for (j = 0; j < n; j++) {
                if (refl[i][j] == UNDEFZ)
                    Rast_set_f_null_value(cell10 + j, 1);
                else
                    cell10[j] = (FCELL) refl[i][j];
            }
            Rast_put_f_row(fd10, cell10);
        }

    }
    if (beam_rad != NULL)
        Rast_close(fd8);
    if (diff_rad != NULL)
        Rast_close(fd9);
    if (refl_rad != NULL)
        Rast_close(fd10);
    Rast_close(fd12);
    return 1;
}




int min(arg1, arg2)
     int arg1;

     int arg2;
{
    int res;

    if (arg1 <= arg2) {
        res = arg1;
    }
    else {
        res = arg2;
    }
    return res;
}

int max(arg1, arg2)
     int arg1;

     int arg2;
{
    int res;

    if (arg1 >= arg2) {
        res = arg1;
    }
    else {
        res = arg2;
    }
    return res;
}



void joules2(struct SunGeometryConstDay *sunGeom,
             struct SunGeometryVarDay *sunVarGeom,
             struct SunGeometryVarSlope *sunSlopeGeom,
             struct SolarRadVar *sunRadVar,
             struct GridGeometry *gridGeom,
             unsigned char *horizonpointer, double latitude, double longitude)
{

    double s0, dfr, dfr_rad;

    double ra, dra;

    double firstTime;

    double firstAngle, lastAngle;

    int ss = 1;

    int srStepNo;

    beam_e = 0.;
    diff_e = 0.;
    refl_e = 0.;
    insol_t = 0.;



    /* all-day radiation */

    srStepNo = (int)(sunGeom->sunrise_time / step);

    if ((sunGeom->sunrise_time - srStepNo * step) > 0.5 * step) {
        firstTime = (srStepNo + 1.5) * step;
    }
    else {
        firstTime = (srStepNo + 0.5) * step;
    }


    firstAngle = (firstTime - 12) * HOURANGLE;
    lastAngle = (sunGeom->sunset_time - 12) * HOURANGLE;




    dfr_rad = step * HOURANGLE;

    sunGeom->timeAngle = firstAngle;

    varCount_global = 0;


    dfr = step;

    while (ss == 1) {

        com_par(sunGeom, sunVarGeom, gridGeom, latitude, longitude);
        s0 = lumcline2(sunGeom, sunVarGeom, sunSlopeGeom, gridGeom,
                       horizonpointer);

        if (sunVarGeom->solarAltitude > 0.) {

            if ((!sunVarGeom->isShadow) && s0 > 0.) {
                insol_t += dfr;
                ra = brad(s0, &bh, sunVarGeom, sunSlopeGeom, sunRadVar);
                beam_e += dfr * ra;
                ra = 0.;
            }
            else
                bh = 0.;
            if (diff_rad != NULL) {
                dra = drad(s0, bh, &rr, sunVarGeom, sunSlopeGeom, sunRadVar);
                diff_e += dfr * dra;
                dra = 0.;
            }
            if (refl_rad != NULL) {
                if (diff_rad == NULL)
                    drad(s0, bh, &rr, sunVarGeom, sunSlopeGeom, sunRadVar);
                refl_e += dfr * rr;
                rr = 0.;
            }
        }                       /* illuminated */


        sunGeom->timeAngle = sunGeom->timeAngle + dfr_rad;

        if (sunGeom->timeAngle > lastAngle) {
            ss = 0;             /* we've got the sunset */
        }
    }                           /* end of while */

}

/*////////////////////////////////////////////////////////////////////// */



void where_is_point(double *length, struct SunGeometryVarDay *sunVarGeom,
                    struct GridGeometry *gridGeom)
{
    double sx, sy;

    double dx, dy;

    double adx, ady;

    int i, j;

    sx = gridGeom->xx0 * invstepx + TOLER;
    sy = gridGeom->yy0 * invstepy + TOLER;

    /*              sx += TOLER; sy += TOLER;
     */
    i = (int)sx;
    j = (int)sy;
    if (i < n - 1 && j < m - 1) {
        /*        zp  = z[j][i]; */

        dx = xx0 - (double)i *gridGeom->stepx;

        dy = yy0 - (double)j *gridGeom->stepy;

        adx = fabs(dx);
        ady = fabs(dy);

        *length = DISTANCE1(gridGeom->xg0, dx, gridGeom->yg0, dy);      /* dist from orig. grid point to the current grid point */

        sunVarGeom->zp = z[j][i];


    }
}







/*////////////////////////////////////////////////////////////////////// */

void calculate(double zmax, double singleLinke, double singleAlbedo,
               struct SolarRadVar globalRadValues,
               struct GridGeometry gridGeom)
{
    int undefval;

    int i, j, k, l;
    int daysInMonth[12] = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
    int mon;

    int arrayOffset = 0;

    int optimum;

    int numRows;

    int horizonOffset = 0;

    double latid_l, cos_u, cos_v, sin_u, sin_v;

    double sin_phi_l, tan_lam_l;

    double latitude, longitude;

    double aspect;

    double slope;

    double q1;

    double slopestart, slopelast = 0.;

    double slopediff = M_PI / 12.;

    double onedeg = M_PI / 180;

    double totrad, optrad;

    double optslope;

    float optbeam_e, optdiff_e, optrefl_e;

    float optb2, optd2, optr2;

    struct SunGeometryConstDay sunGeom[12];

    struct SunGeometryVarDay sunVarGeom[12];

    struct SunGeometryVarSlope sunSlopeGeom[12];

    struct SolarRadVar sunRadVar[12];

    /*
       BeamRadFunc bRadFunc;
       DiffRadFunc dRadFunc;
     */
    numRows = m / numPartitions;

    fprintf(stderr, "\n\n");

    /* Set up parameters for projection to lat/long  */

    if ((in_proj_info = G_get_projinfo()) == NULL)
        G_fatal_error
            ("Can't get projection info of current location: \n please set latitude via 'lat' or 'latin' option!");

    if ((in_unit_info = G_get_projunits()) == NULL)
        G_fatal_error("Can't get projection units of current location");

    if (pj_get_kv(&iproj, in_proj_info, in_unit_info) < 0)
        G_fatal_error("Can't get projection key values of current location");


    for (i = 0; i < 12; i++) {
        sunVarGeom[i].zmax = zmax;
        sunGeom[i].sindecl = sin(declination[i]);
        sunGeom[i].cosdecl = cos(declination[i]);
        sunRadVar[i].G_norm_extra = com_sol_const(day[i]);
    }



    if (beam_rad != NULL) {
        beam = (float **)malloc(sizeof(float *) * (m));
        for (l = 0; l < m; l++) {
            beam[l] = (float *)malloc(sizeof(float) * (n));
        }

        for (j = 0; j < m; j++) {
            for (i = 0; i < n; i++)
                beam[j][i] = UNDEFZ;
        }
    }


    if (opt_angle != NULL) {
        optang = (float **)malloc(sizeof(float *) * (m));
        for (l = 0; l < m; l++) {
            optang[l] = (float *)malloc(sizeof(float) * (n));
        }

        for (j = 0; j < m; j++) {
            for (i = 0; i < n; i++)
                optang[j][i] = UNDEFZ;
        }
    }


    if (diff_rad != NULL) {
        diff = (float **)malloc(sizeof(float *) * (m));
        for (l = 0; l < m; l++) {
            diff[l] = (float *)malloc(sizeof(float) * (n));
        }

        for (j = 0; j < m; j++) {
            for (i = 0; i < n; i++)
                diff[j][i] = UNDEFZ;
        }
    }

    if (refl_rad != NULL) {
        refl = (float **)malloc(sizeof(float *) * (m));
        for (l = 0; l < m; l++) {
            refl[l] = (float *)malloc(sizeof(float) * (n));
        }

        for (j = 0; j < m; j++) {
            for (i = 0; i < n; i++)
                refl[j][i] = UNDEFZ;
        }
    }

    if (linkein == NULL) {
        for (mon = 0; mon < 12; mon++) {
            sunRadVar[mon].linke = singleLinke;
        }

    }
    if (albedo == NULL) {
        for (mon = 0; mon < 12; mon++) {
            sunRadVar[mon].alb = singleAlbedo;
        }

    }

    if (coefbh == NULL) {
        for (mon = 0; mon < 12; mon++) {
            sunRadVar[mon].cbh = BSKY;
        }
    }
    if (coefdh == NULL) {
        for (mon = 0; mon < 12; mon++) {
            sunRadVar[mon].cdh = DSKY;
        }
    }

    for (j = 0; j < m; j++) {
        if (j % numRows == 0) {
            INPUT_part(j);
            arrayOffset = 0;
            horizonOffset = 0;

        }


        if (latin != NULL) {
            slopelast = la[j][0] * deg2rad;
        }

        G_percent(j, m - 1, 2);
        for (i = 0; i < n; i++) {


            gridGeom.xg0 = gridGeom.xx0 = (double)i *gridGeom.stepx;

            gridGeom.xp = xmin + gridGeom.xx0;
            gridGeom.yg0 = gridGeom.yy0 = (double)j *gridGeom.stepy;

            gridGeom.yp = ymin + gridGeom.yy0;
            func = NULL;

            for (mon = 0; mon < 12; mon++) {
                sunVarGeom[mon].z_orig = sunVarGeom[mon].zp = z[j][i];

            }

            if (sunVarGeom[0].z_orig != UNDEFZ) {
                if (o[j][i] != 0.)
                    aspect = o[j][i] * deg2rad;
                else
                    aspect = UNDEF;
                if (linkein != NULL) {
                    sunRadVar[0].linke = li1[arrayOffset][i];
                    sunRadVar[1].linke = li2[arrayOffset][i];
                    sunRadVar[2].linke = li3[arrayOffset][i];
                    sunRadVar[3].linke = li4[arrayOffset][i];
                    sunRadVar[4].linke = li5[arrayOffset][i];
                    sunRadVar[5].linke = li6[arrayOffset][i];
                    sunRadVar[6].linke = li7[arrayOffset][i];
                    sunRadVar[7].linke = li8[arrayOffset][i];
                    sunRadVar[8].linke = li9[arrayOffset][i];
                    sunRadVar[9].linke = li10[arrayOffset][i];
                    sunRadVar[10].linke = li11[arrayOffset][i];
                    sunRadVar[11].linke = li12[arrayOffset][i];
                }
                if (albedo != NULL) {
                    for (mon = 0; mon < 12; mon++) {
                        sunRadVar[mon].alb = a[j][i];
                    }
                }
                if (latin != NULL) {
                    latitude = la[j][i];
                    latitude = latitude * deg2rad;
                }
                if (latin == NULL && lt == NULL) {
                    if ((G_projection() != PROJECTION_LL)) {

                        longitude = gridGeom.xp;
                        latitude = gridGeom.yp;



                        /*

                           pj_zero_proj(&oproj);
                         */
                        sprintf(oproj.proj, "%s", "ll");


                        if (pj_do_proj(&longitude, &latitude, &iproj, &oproj)
                            < 0) {
                            fprintf(stderr, "Error in pj_do_proj\n");
                            exit(0);
                        }

                        la_max = amax1(la_max, latitude);
                        la_min = amin1(la_min, latitude);
                        latitude *= deg2rad;
                        longitude *= deg2rad;
                    }
                    else {      /* ll projection */
                        longitude = gridGeom.xp;
                        latitude = gridGeom.yp;
                        la_max = amax1(la_max, latitude);
                        la_min = amin1(la_min, latitude);
                        latitude *= deg2rad;
                        longitude *= deg2rad;
                    }
                }

                if (coefbh != NULL) {
                    sunRadVar[0].cbh = cbhr1[arrayOffset][i];
                    sunRadVar[1].cbh = cbhr2[arrayOffset][i];
                    sunRadVar[2].cbh = cbhr3[arrayOffset][i];
                    sunRadVar[3].cbh = cbhr4[arrayOffset][i];
                    sunRadVar[4].cbh = cbhr5[arrayOffset][i];
                    sunRadVar[5].cbh = cbhr6[arrayOffset][i];
                    sunRadVar[6].cbh = cbhr7[arrayOffset][i];
                    sunRadVar[7].cbh = cbhr8[arrayOffset][i];
                    sunRadVar[8].cbh = cbhr9[arrayOffset][i];
                    sunRadVar[9].cbh = cbhr10[arrayOffset][i];
                    sunRadVar[10].cbh = cbhr11[arrayOffset][i];
                    sunRadVar[11].cbh = cbhr12[arrayOffset][i];
                }
                if (coefdh != NULL) {
                    sunRadVar[0].cdh = cdhr1[arrayOffset][i];
                    sunRadVar[1].cdh = cdhr2[arrayOffset][i];
                    sunRadVar[2].cdh = cdhr3[arrayOffset][i];
                    sunRadVar[3].cdh = cdhr4[arrayOffset][i];
                    sunRadVar[4].cdh = cdhr5[arrayOffset][i];
                    sunRadVar[5].cdh = cdhr6[arrayOffset][i];
                    sunRadVar[6].cdh = cdhr7[arrayOffset][i];
                    sunRadVar[7].cdh = cdhr8[arrayOffset][i];
                    sunRadVar[8].cdh = cdhr9[arrayOffset][i];
                    sunRadVar[9].cdh = cdhr10[arrayOffset][i];
                    sunRadVar[10].cdh = cdhr11[arrayOffset][i];
                    sunRadVar[11].cdh = cdhr12[arrayOffset][i];
                }

                undefval = 0;

                for (k = 0; k < 12; k++) {
                    if ((sunRadVar[k].cbh == UNDEFZ)
                        || (sunRadVar[k].cdh == UNDEFZ)) {
                        undefval = 1;
                        break;
                    }
                }

                if (undefval) {
                    if (beam_rad != NULL)
                        beam[j][i] = UNDEFZ;
                    if (diff_rad != NULL)
                        diff[j][i] = UNDEFZ;
                    if (refl_rad != NULL)
                        refl[j][i] = UNDEFZ;
                    optang[j][i] = UNDEFZ;

                    /* Continue with the next raster point */

                    undefval = 0;

                    continue;
                }
                for (k = 0; k < 12; k++) {
                    if (sunRadVar[k].cbh < 0.)
                        sunRadVar[k].cbh = 0.01;
                    if (sunRadVar[k].cdh < 0.)
                        sunRadVar[k].cdh = 0.01;

                }
                gridGeom.sinlat = sin(-latitude);
                gridGeom.coslat = cos(-latitude);


                cos_v = cos(M_PI / 2 + aspect);
                sin_v = sin(M_PI / 2 + aspect);
                for (mon = 0; mon < 12; mon++) {
                    sunSlopeGeom[mon].aspect = aspect;
                }
                /* Reset onedeg to positive value */
                onedeg = fabs(onedeg);

                if (i == 0) {
                    if (latitude > 0) {
                        slopestart = latitude - slopediff;
                    }
                    else {
                        slopestart = -latitude - slopediff;
                    }
                }
                else {
                    slopestart = slopelast;
                }

                /* Start at nearest whole degree. */
                slopestart *= rad2deg;
                slopestart = (int)(slopestart);
                slopestart *= deg2rad;

                optslope = slope = slopestart;
                optimum = 0;
                optrad = 0.;

                /* Calculate first value of total radiation */


                cos_u = cos(M_PI / 2 - slope);
                sin_u = sin(M_PI / 2 - slope);

                sin_phi_l = -gridGeom.coslat * cos_u * sin_v
                    + gridGeom.sinlat * sin_u;
                latid_l = asin(sin_phi_l);

                q1 = gridGeom.sinlat * cos_u * sin_v +
                    gridGeom.coslat * sin_u;
                tan_lam_l = -cos_u * cos_v / q1;
                totrad = 0;
                optrad = optbeam_e = optdiff_e = optrefl_e = 0.;
                for (mon = 0; mon < 12; mon++) {
                    com_par_const(0., sunGeom + mon, &gridGeom);
                    sunSlopeGeom[mon].slope = slope;
                    sunSlopeGeom[mon].longit_l = atan(tan_lam_l);
                    sunSlopeGeom[mon].lum_C31_l =
                        cos(latid_l) * sunGeom[mon].cosdecl;
                    sunSlopeGeom[mon].lum_C33_l =
                        sin_phi_l * sunGeom[mon].sindecl;
                    joules2(sunGeom + mon, sunVarGeom + mon,
                            sunSlopeGeom + mon, sunRadVar + mon, &gridGeom,
                            horizonarray + horizonOffset, latitude,
                            longitude);

                    optbeam_e += beam_e * daysInMonth[mon];
                    optdiff_e += diff_e * daysInMonth[mon];
                    optrefl_e += refl_e * daysInMonth[mon];
                    optrad += daysInMonth[mon] * (beam_e + diff_e + refl_e);
                }
                /* Calculate second value of total radiation */

                slope += onedeg;
                cos_u = cos(M_PI / 2 - slope);
                sin_u = sin(M_PI / 2 - slope);

                sin_phi_l = -gridGeom.coslat * cos_u * sin_v
                    + gridGeom.sinlat * sin_u;
                latid_l = asin(sin_phi_l);

                q1 = gridGeom.sinlat * cos_u * sin_v
                    + gridGeom.coslat * sin_u;
                tan_lam_l = -cos_u * cos_v / q1;

                optb2 = optd2 = optr2 = 0.;
                for (mon = 0; mon < 12; mon++) {

                    sunSlopeGeom[mon].slope = slope;
                    sunSlopeGeom[mon].longit_l = atan(tan_lam_l);
                    sunSlopeGeom[mon].lum_C31_l =
                        cos(latid_l) * sunGeom[mon].cosdecl;
                    sunSlopeGeom[mon].lum_C33_l =
                        sin_phi_l * sunGeom[mon].sindecl;

                    joules2(sunGeom + mon, sunVarGeom + mon,
                            sunSlopeGeom + mon,
                            sunRadVar + mon, &gridGeom,
                            horizonarray + horizonOffset,
                            latitude, longitude);
                    totrad += daysInMonth[mon] * (beam_e + diff_e + refl_e);
                    optb2 += beam_e * daysInMonth[mon];
                    optd2 += diff_e * daysInMonth[mon];
                    optr2 += refl_e * daysInMonth[mon];
                }




                if (totrad < optrad) {
                    /* Search in the other direction */
                    slope -= onedeg;

                    onedeg *= -1.;
                }
                else {
                    optbeam_e = optb2;
                    optdiff_e = optd2;
                    optrefl_e = optr2;
                    optrad = totrad;
                    optslope = slope;
                }


                while (!optimum) {

                    slope += onedeg;
                    cos_u = cos(M_PI / 2 - slope);
                    sin_u = sin(M_PI / 2 - slope);


                    sin_phi_l = -gridGeom.coslat * cos_u * sin_v
                        + gridGeom.sinlat * sin_u;
                    latid_l = asin(sin_phi_l);

                    q1 = gridGeom.sinlat * cos_u * sin_v
                        + gridGeom.coslat * sin_u;
                    tan_lam_l = -cos_u * cos_v / q1;

                    totrad = 0.;
                    optb2 = optd2 = optr2 = 0.;
                    for (mon = 0; mon < 12; mon++) {
                        sunSlopeGeom[mon].slope = slope;
                        sunSlopeGeom[mon].longit_l = atan(tan_lam_l);
                        sunSlopeGeom[mon].lum_C31_l =
                            cos(latid_l) * sunGeom[mon].cosdecl;
                        sunSlopeGeom[mon].lum_C33_l =
                            sin_phi_l * sunGeom[mon].sindecl;
                        joules2(sunGeom + mon, sunVarGeom + mon,
                                sunSlopeGeom + mon, sunRadVar + mon,
                                &gridGeom, horizonarray + horizonOffset,
                                latitude, longitude);
                        totrad +=
                            daysInMonth[mon] * (beam_e + diff_e + refl_e);
                        optb2 += daysInMonth[mon] * beam_e;
                        optd2 += daysInMonth[mon] * diff_e;
                        optr2 += daysInMonth[mon] * refl_e;

                    }

                    if (totrad > optrad) {
                        optbeam_e = optb2;
                        optdiff_e = optd2;
                        optrefl_e = optr2;
                        optrad = totrad;
                        optslope = slope;
                    }
                    else {
                        /* We have found the optimal angle */
                        optimum = 1;
                    }
                }               /* End while !optimum */

                if ((beam_rad != NULL)
                    || (diff_rad != NULL) || (refl_rad != NULL)) {
                    if (beam_rad != NULL)
                        beam[j][i] = (float)optbeam_e / 365.;
                    if (diff_rad != NULL)
                        diff[j][i] = (float)optdiff_e / 365.;
                    if (refl_rad != NULL)
                        refl[j][i] = (float)optrefl_e / 365.;
                }
                /*
                   if(optslope<0.)
                   {
                   optang[j][i] = 0.;
                   }
                   else
                   {                    
                 */
                optang[j][i] = optslope * rad2deg;
                /*
                   }
                 */

                slopelast = optslope;

            }                   /* undefs */
            horizonOffset += arrayNumInt;
        }                       /* End loop over columns */

        arrayOffset++;

    }
    fprintf(stderr, "\n");

}



double com_declin(int no_of_day)
{
    double d1, decl;

    d1 = pi2 * no_of_day / 365.25;
    decl = asin(0.3978 * sin(d1 - 1.4 + 0.0355 * sin(d1 - 0.0489)));
    decl = -decl;
    /*      printf(" declination : %lf\n", decl); */

    return (decl);
}
